// FileManager.cpp: FileManager NX̃Cve[V
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MeshEditor.h"
#include "FileManager.h"

#include "MeshData.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// \z/
//////////////////////////////////////////////////////////////////////

FileManager::FileManager()
{

}

FileManager::~FileManager()
{

}

void FileManager::open(MeshData *mesh)
{
	//read cloud data file and perform trianlization.
	if(file_ext == "data"){
		this->openCloud(mesh);
	}
	//read vertex data from ply file
	else if(file_ext == "ply"){
		this->openPly(mesh);
	}
	//read vertex data from ply2 file(my original format)
	else if(file_ext == "ply2"){
		this->openOriginal(mesh);
	}
	//read vertex data from obj file(may be wave-front file)
	else if(file_ext == "obj"){
		this->openWaveFront(mesh);
	}
	else if(file_ext == "plyR"){
		this->openRidge(mesh);
	}
	mesh->normalizeScale();
	mesh->generateVertexLink();
	mesh->generateFaceLink();
}

void FileManager::setFile(FILE *file, CString file_name, CString file_ext)
{
	this->file =  file;
	this->file_name = file_name;
	this->file_ext = file_ext;
}

void FileManager::openCloud(MeshData *mesh)
{
	int vertex_N;
	fscanf(file, "%d", &vertex_N);

	mesh->setVertexCount(vertex_N);
	float p[3];
	for(int i=0; i<vertex_N; i++){
		fscanf(file, "%f", &p[0]);
		fscanf(file, "%f", &p[1]);
		fscanf(file, "%f", &p[2]);
		mesh->setVertex(i, p);
	}
	//generateFaceFromPoints();
}

void FileManager::openPly(MeshData *mesh)
{
	int vertex_N, face_N;

	fscanf(file, "%d", &vertex_N);
	mesh->setVertexCount(vertex_N);

	fscanf(file, "%d", &face_N);
	mesh->setFaceCount(face_N);

	float p[3];
	float dummy;
	for(int i=0; i<vertex_N; i++){
		fscanf(file, "%f", &p[0]);
		fscanf(file, "%f", &p[2]);
		fscanf(file, "%f", &p[1]);
		mesh->setVertex(i, p);

		//skip extra data
		fscanf(file, "%f", &dummy);
		fscanf(file, "%f", &dummy);
	}

	int f[3];
	for(i=0; i<face_N; i++){
		fscanf(file, "%f", &dummy);
		fscanf(file, "%d", &f[0]);
		fscanf(file, "%d", &f[1]);
		fscanf(file, "%d", &f[2]);
		mesh->setFace(i, f);
	}
}

void FileManager::openOriginal(MeshData *mesh)
{
	int vertex_N, face_N;

	fscanf(file, "%d", &vertex_N);
	mesh->setVertexCount(vertex_N);

	fscanf(file, "%d", &face_N);
	mesh->setFaceCount(face_N);

	float p[3];
	float dummy;
	for(int i=0; i<vertex_N; i++){
		fscanf(file, "%f", &p[0]);
		fscanf(file, "%f", &p[1]);
		fscanf(file, "%f", &p[2]);
		//p[0] *= -1;
		mesh->setVertex(i, p);
	}

	int f[3];
	for(i=0; i<face_N; i++){
		fscanf(file, "%f", &dummy);
		fscanf(file, "%d", &f[0]);
		fscanf(file, "%d", &f[1]);
		fscanf(file, "%d", &f[2]);
		mesh->setFace(i, f);
	}
}

void FileManager::openWaveFront(MeshData *mesh)
{
	/*
	char dummyS[10];
		fscanf(file, "%s", dummyS);
		fscanf(file, "%s", dummyS);
		fscanf(file, "%d", &(mesh->vertex_N));
		fscanf(file, "%s", dummyS);

		fscanf(file, "%s", dummyS);
		fscanf(file, "%s", dummyS);
		fscanf(file, "%s", dummyS);
		fscanf(file, "%d", &(mesh->face_N));
		fscanf(file, "%s", dummyS);

		mesh->vertex = new float[mesh->vertex_N][3];
		*/
}

void FileManager::openRidge(MeshData *mesh)
{
	int vertex_N, face_N;

	fscanf(file, "%d", &vertex_N);
	mesh->setVertexCount(vertex_N);

	fscanf(file, "%d", &face_N);
	mesh->setFaceCount(face_N);

	float p[3];
	float dummy;
	for(int i=0; i<vertex_N; i++){
		fscanf(file, "%f", &p[0]);
		fscanf(file, "%f", &p[1]);
		fscanf(file, "%f", &p[2]);
		//p[0] *= -1;
		mesh->setVertex(i, p);
	}

	int f[3];
	for(i=0; i<face_N; i++){
		fscanf(file, "%f", &dummy);
		fscanf(file, "%d", &f[0]);
		fscanf(file, "%d", &f[1]);
		fscanf(file, "%d", &f[2]);
		mesh->setFace(i, f);
	}

	BOOL* isRidge = mesh->isRidge = new BOOL[vertex_N];
	BOOL* isRavine = mesh->isRavine = new BOOL[vertex_N];
	for(i=0; i<vertex_N ; i++){
		isRidge[i] = false;
		isRavine[i] = false;
	}

	int ridge_N;
	fscanf(file, "%d", &ridge_N);
	int ravine_N;
	fscanf(file, "%d", &ravine_N);
	for(i=0; i<ridge_N; i++){
		int index;
		fscanf(file, "%d", &index);
		isRidge[index] = true;
	}
	for(i=0; i<ravine_N; i++){
		int index;
		fscanf(file, "%d", &index);
		isRavine[index] = true;
	}
}

void FileManager::save(MeshData *mesh)
{
	if(file_ext == "ply2"){
		this->saveOriginal(mesh);
	}
	else if(file_ext == "plyR"){
		this->saveRidge(mesh);
	}
}

void FileManager::saveOriginal(MeshData *mesh)
{
	int f_N = mesh->countValidFace();
	int v_N = mesh->countValidVertex();
	int* vertex_table = new int[mesh->vertex_N];
	int counter = 0;

	float (*vertex)[3] = mesh->vertex;
	int (*face)[3] = mesh->face;
	int vertex_N = mesh->vertex_N;
	int face_N = mesh->face_N;

	fprintf(file, "%d\n", v_N);
	fprintf(file, "%d\n", f_N);
	for(int i=0; i<vertex_N; i++){
		if(mesh->degree_v[i] != 0){
			fprintf(file, "%f ", vertex[i][0]);
			fprintf(file, "%f ", vertex[i][1]);
			fprintf(file, "%f\n", vertex[i][2]);
			vertex_table[i] = counter;
			counter++;
		}
		else
			vertex_table[i] = -1;
	}
	for(i=0; i<face_N; i++){
		if(mesh->face[i][0] >= 0){
			fprintf(file, "3 ");
			fprintf(file, "%d ", vertex_table[face[i][0]]);
			fprintf(file, "%d ", vertex_table[face[i][1]]);
			fprintf(file, "%d\n", vertex_table[face[i][2]]);
		}
	}
	delete[] vertex_table;
}

void FileManager::saveRidge(MeshData *mesh)
{
	int f_N = mesh->countValidFace();
	int v_N = mesh->countValidVertex();
	int* vertex_table = new int[mesh->vertex_N];
	int counter = 0;

	float (*vertex)[3] = mesh->vertex;
	int (*face)[3] = mesh->face;
	int vertex_N = mesh->vertex_N;
	int face_N = mesh->face_N;

	fprintf(file, "%d\n", v_N);
	fprintf(file, "%d\n", f_N);
	for(int i=0; i<vertex_N; i++){
		if(mesh->degree_v[i] != 0){
			fprintf(file, "%f ", vertex[i][0]);
			fprintf(file, "%f ", vertex[i][1]);
			fprintf(file, "%f\n", vertex[i][2]);
			vertex_table[i] = counter;
			counter++;
		}
		else
			vertex_table[i] = -1;
	}
	for(i=0; i<face_N; i++){
		if(mesh->face[i][0] >= 0){
			fprintf(file, "3 ");
			fprintf(file, "%d ", vertex_table[face[i][0]]);
			fprintf(file, "%d ", vertex_table[face[i][1]]);
			fprintf(file, "%d\n", vertex_table[face[i][2]]);
		}
	}

	int ridge_N = 0;
	int ravine_N = 0;
	for(i=0; i<vertex_N; i++){
		if(mesh->isRidge[i])
			ridge_N++;
		if(mesh->isRavine[i])
			ravine_N++;
	}
	fprintf(file, "%d\n", ridge_N);
	fprintf(file, "%d\n", ravine_N);
	for(i=0; i<vertex_N; i++)
		if(mesh->isRidge[i])
			fprintf(file, "%d\n", vertex_table[i]);
	for(i=0; i<vertex_N; i++)
		if(mesh->isRavine[i])
			fprintf(file, "%d\n", vertex_table[i]);

	delete[] vertex_table;
}
